var proauth = (function () {
'use strict';

var absolutePath = (href) => {
  const link = document.createElement("a");
  link.href = href;
  return `${link.protocol}//${link.host}${link.pathname}${link.search}${link.hash}`
};

const originalFetch$1 = window.fetch;

const fetch$1 = function (input, init) {
  return new Promise((resolve) => {
    const request = new Request(input, init);

    if (originalFetch$1.polyfill) {
      // fetch polyfill does not absolutize the url, so...
      request.url = absolutePath(request.url);
    }

    const fetchEvent = new CustomEvent("fetch");
    fetchEvent.request = request;
    fetchEvent.respondWith = resolve;
    this.dispatchEvent(fetchEvent); // eslint-disable-line no-invalid-this

    // (todo): when to reject?
  })
};

var fetch$2 = {
  install: (observer) => {
    window.fetch = fetch$1.bind(observer);
  },
  uninstall: () => {
    window.fetch = originalFetch$1;
  },
  fetch: originalFetch$1
};

const originalXMLHttpRequest = window.XMLHttpRequest;

const _readyState = Symbol("readyState");
const _responseBody = Symbol("responseBody");
const _response = Symbol("response");
const _requestHeaders = Symbol("requestHeaders");
const _method = Symbol("method");
const _url = Symbol("url");
const _username = Symbol("username");
const _password = Symbol("password");
const _sendFlag = Symbol("sendFlag");
const _errorFlag = Symbol("errorFlag");
const _forceMimeType = Symbol("forceMimeType");
const _responseHeaders = Symbol("responseHeaders");
const _aborted = Symbol("aborted");
const _withCredentials = Symbol("withCredentials");

const responseParser = {
  "": "text",
  "arraybuffer": "arrayBuffer",
  "blob": "blob",
  "json": "json",
  "text": "text"
};

/*
  Without mocking, the native XMLHttpRequest object will throw
  an error when attempting to set these headers. We match this behavior.
*/
const unsafeHeaders = {
  "accept-charset": true,
  "accept-encoding": true,
  "connection": true,
  "content-length": true,
  "cookie": true,
  "cookie2": true,
  "content-transfer-encoding": true,
  "date": true,
  "expect": true,
  "host": true,
  "keep-alive": true,
  "referer": true,
  "te": true,
  "trailer": true,
  "transfer-encoding": true,
  "upgrade": true,
  "user-agent": true,
  "via": true
};

const throwMethodError = function (method, message) {
  throw new Error(`Failed to execute '${method}' on 'XMLHttpRequest': ${message}.`)
};

const throwReadPropError = function (property, message) {
  throw new Error(`Failed to read the '${property}' property from 'XMLHttpRequest': ${message}.`)
};

const verifyState = function (xhr, method) {
  if (xhr.readyState !== XMLHttpRequest.OPENED) {
    throwMethodError(method, "The object's state must be OPENED");
  }

  if (xhr[_sendFlag]) {
    throwMethodError(method, "Already sent");
  }
};

/*
  Cross-browser XML parsing. Used to turn
  XML responses into Document objects
  Borrowed from JSpec
*/
const parseXML = function (text) {
  let xmlDoc = null;

  if (typeof DOMParser !== "undefined") {
    const parser = new DOMParser();
    xmlDoc = parser.parseFromString(text, "text/xml");

    // Return an XML parse error to get namespace of "parsererror" tag
    const parsererrorNS = parser.parseFromString("INVALID", "text/xml")
      .getElementsByTagName("parsererror")[0].namespaceURI;

    // Check if xmlDoc resulted in an XML parse error
    if (xmlDoc.getElementsByTagNameNS(parsererrorNS, "parsererror").length > 0) {
      return null
    }
  } else {
    xmlDoc = new ActiveXObject("Microsoft.XMLDOM"); // eslint-disable-line no-undef
    xmlDoc.async = "false";
    xmlDoc.loadXML(text);
  }

  return xmlDoc
};

/**
 * Places a XMLHttpRequest object into the passed state.
 *
 * @param {XMLHttpRequest} xhr The XHR object
 * @param {string} state The ready state
 * @return {void}
 */
const readyStateChange = function (xhr, state) {
  xhr[_readyState] = state;
  xhr.dispatchEvent(new Event("readystatechange"));
};

class XMLHttpRequestToFetch extends XMLHttpRequest {

  get readyState() {
    return this[_readyState] || this.UNSENT
  }

  get status() {
    return this[_response] ? this[_response].status : 0
  }

  get statusText() {
    return this[_response] ? this[_response].statusText : ""
  }

  get response() {
    // https://xhr.spec.whatwg.org/#xmlhttprequest-response
    const type = this.responseType;
    const state = this[_readyState];
    if (type === "" || type === "text") {
      return state !== this.LOADING && state !== this.DONE ? "" : this[_responseBody]
    }

    if (state !== this.DONE) {
      return null
    }

    return this[_responseBody]

  }

  get responseURL() {
    if (this[_response] && this.readyState === this.DONE) {
      return this[_response].url
    }

    return ""
  }

  get responseText() {
    const responseType = this.responseType;
    if (responseType !== "" && responseType !== "text") {
      throwReadPropError(
        "responseType",
        `The value is only accessible if the object's 'responseType' is '' ` +
        `or 'text' (was '${responseType}')`
      );
    }

    const state = this[_readyState];
    if (state !== this.LOADING && state !== this.DONE) {
      return ""
    }

    return this[_responseBody] || ""
  }

  get responseXML() {
    const responseType = this.responseType;
    if (responseType !== "" && responseType !== "document") {
      throwReadPropError(
        "responseXML",
        `The value is only accessible if the object's 'responseType' is '' ` +
        `or 'document' (was '${responseType}')`);
    }

    const body = this[_responseBody];

    if (this[_readyState] !== this.DONE || !!body) {
      return null
    }

    const finalMimeType = this.getResponseHeader("Content-Type");

    if (finalMimeType !== null &&
      !/(text\/html)|(text\/xml)|(application\/xml)|(\+xml)/i.test(finalMimeType)) {
      return null
    }

    if (responseType === "" && finalMimeType === "text/html") {
      return null
    }

    return parseXML(body)
  }

  get withCredentials() {
    return this[_withCredentials] || false
  }

  set withCredentials(enable) {
    this[_withCredentials] = !!enable;
  }

  /**
   * Duplicates the behavior of native XMLHttpRequest's open function
   *
   * @param {string} method Method of the request
   * @param {string} url URL of the request
   * @param {boolean} async Wheter to make the request synchronously or not
   * @param {string} username Username to use in the request
   * @param {string} password Password to use in the request
   * @returns {void}
   */
  open(method, url, async = true, username, password) { // eslint-disable-line max-params

    if (method === undefined || url === undefined) {
      throw new TypeError("Not enough arguments to XMLHttpRequest.open")
    }

    if (!async) {
      if (console && console.warn) { console.warn("Synchronous XHR are not supported"); }
    }

    this[_method] = method;
    this[_url] = url;
    this[_username] = username;
    this[_password] = password;
    this[_requestHeaders] = {};
    this[_sendFlag] = false;
    readyStateChange(this, this.OPENED);
  }

  /**
   * Duplicates the behavior of native XMLHttpRequest's setRequestHeader function
   *
   * @param {string} header Header key
   * @param {string} value Header value
   * @returns {void}
   */
  setRequestHeader(header, value) {
    verifyState(this, "setRequestHeader");

    if (unsafeHeaders[String(header).toLowerCase()] || /^(Sec-|Proxy-)/i.test(header)) {
      console.error(`Refused to set unsafe header '${header}'`);
      return
    }

    if (this[_requestHeaders][header]) {
      this[_requestHeaders][header] += `,${value}`;
    } else {
      this[_requestHeaders][header] = value;
    }
  }

  /**
   * Duplicates the behavior of native XMLHttpRequest's send function
   *
   * @param {object|null} data Request body
   * @returns {void}
   */
  send(data) {
    verifyState(this, "send");
    this[_errorFlag] = false;
    this[_sendFlag] = true;

    this.dispatchEvent(new ProgressEvent("loadstart", { bubbles: false, cancelable: false }));
    //this.dispatchEvent(new ProgressEvent("progress", { bubbles: false, cancelable: false }))

    const fetchInit = {
      method: this[_method],
      headers: this[_requestHeaders],
      body: data
      // mode: // NOT supported by xhr
      // credentials: // TODO
    };

    if (this[_withCredentials]) {
      fetchInit.credentials = "include";
    }

    const p = Promise.race([
      fetch(this[_url], fetchInit).then(
        (response) => response,
        (reason) => new Promise((resolve, reject) => reject(reason))
      ),
      new Promise((resolve, reject) => {
        if (this.timeout > 0) {
          setTimeout(() => reject(new Error("request timeout")), this.timeout);
        }
      })
    ]);

    p.then(
      (response) => {
        // fetch resolved

        this[_response] = response;

        // HEADERS_RECEIVED Stage
        const headers = response.headers;
        this[_responseHeaders] = {};
        let mimeTypeForced = false;

        headers.forEach((value, key) => {
          if (this[_forceMimeType] && key.toLowerCase() === "content-type") {
            this[_responseHeaders][key] = this[_forceMimeType];
            mimeTypeForced = true;
          } else {
            this[_responseHeaders][key] = value;
          }
        });

        if (this[_forceMimeType] && !mimeTypeForced) {
          this[_responseHeaders]["Content-Type"] = this[_forceMimeType];
        }

        readyStateChange(this, this.HEADERS_RECEIVED);

        // LOADING Stage
        readyStateChange(this, this.LOADING);
        this.dispatchEvent(new ProgressEvent("progress", { bubbles: false, cancelable: false }));
        // TODO: do we need body size?

        return response[
          responseParser[this.responseType] || "text"
        ]()
      },
      (reason) => {
        // fetch rejected || request timeout
        readyStateChange(this, this.DONE);

        if (reason instanceof Error && reason.message === "request timeout") {
          // request timeout
          this.dispatchEvent(new ProgressEvent("timeout", { bubbles: false, cancelable: false }));
        } else if (!this[_aborted]) {
          // fetch rejected
          this.dispatchEvent(new ProgressEvent("error", { bubbles: false, cancelable: false }));
        }

        return new Promise((resolve, reject) => reject(reason))
      }
    ).then((body) => {
      // fetch resolved && body is ready
      if (!this[_aborted] && this[_readyState]) {
        this[_responseBody] = body;
        this[_sendFlag] = false;
        readyStateChange(this, this.DONE);
        this.dispatchEvent(new ProgressEvent("load", { bubbles: false, cancelable: false }));
      }
    }).then(
      // finally, fire loadend event when the fetch completed (success or failure)
      () => {
        this.dispatchEvent(new ProgressEvent("loadend", { bubbles: false, cancelable: false }));
      },
      () => {
        this.dispatchEvent(new ProgressEvent("loadend", { bubbles: false, cancelable: false }));
      }
      );
  }

  /**
   * Duplicates the behavior of native XMLHttpRequest's abort function
   *
   * @returns {void}
   */
  abort() {
    this[_aborted] = true;
    this[_errorFlag] = true;
    this[_requestHeaders] = {};

    if ((this[_readyState] === this.OPENED && this[_sendFlag]) ||
      this[_readyState] === this.HEADERS_RECEIVED ||
      this[_readyState] === this.LOADING) {
      readyStateChange(this, this.DONE);
      this[_sendFlag] = false;
    }

    this[_readyState] = this.UNSENT;

    this.dispatchEvent(new ProgressEvent("abort", { bubbles: false, cancelable: false }));
  }

  /**
   * Duplicates the behavior of native XMLHttpRequest's getResponseHeader function
   *
   * @param {string} header Header key
   * @returns {string|null} Header value
   */
  getResponseHeader(header) {
    if (this[_readyState] < this.HEADERS_RECEIVED) {
      return null
    }

    if (/^Set-Cookie2?$/i.test(header)) {
      return null
    }

    header = header.toLowerCase();

    for (const h in this[_responseHeaders]) {
      if (h.toLowerCase() === header) {
        return this[_responseHeaders][h]
      }
    }

    return null
  }

  /**
   * Duplicates the behavior of native XMLHttpRequest's getAllResponseHeaders function
   *
   * @returns {array} List of response headers
   */
  getAllResponseHeaders() {
    if (this[_readyState] < this.HEADERS_RECEIVED) {
      return ""
    }

    let responseHeaders = this[_responseHeaders];
    if (responseHeaders && responseHeaders.length) {
      // Clone headers and sort them (@see https://xhr.spec.whatwg.org/#dom-xmlhttprequest-getallresponseheaders)
      responseHeaders = responseHeaders.slice().sort();
    }
    let headers = "";

    for (const header in responseHeaders) {
      if (responseHeaders.hasOwnProperty(header) && !/^Set-Cookie2?$/i.test(header)) {
        headers += `${header}: ${responseHeaders[header]}\r\n`;
      }
    }

    return headers
  }

  /**
   * Duplicates the behavior of native XMLHttpRequest's overrideMimeType function
   *
   * @param {string} mimeType MimeType that will be used
   * @returns {void}
   */
  overrideMimeType(mimeType) {
    if (typeof mimeType === "string") {
      this[_forceMimeType] = mimeType.toLowerCase();
    }
  }

}

var xhr = {
  install: () => {
    window.XMLHttpRequest = XMLHttpRequestToFetch;
  },
  uninstall: () => {
    window.XMLHttpRequest = originalXMLHttpRequest;
  },
  XMLHttpRequest: originalXMLHttpRequest
};

class EventTarget {

  /**
   * Create an event target.
   */
  constructor() {
    const delegate = document.createDocumentFragment();
    for (const fx of [ "addEventListener", "dispatchEvent", "removeEventListener"]) {
      this[fx] = (...xs) => delegate[fx](...xs);
    }
  }

}

const postMessage = (target, aMessage, transferList) => {

  const ports = [];
  for (const transfer of transferList || []) {
    if (transfer instanceof MessagePort) {
      ports.push(transfer);
    }
  }

  const e = new MessageEvent("message", {
    bubbles: false,
    cancelable: true,
    data: aMessage,
    ports: ports.length > 0 ? ports : undefined
  });

  if (target.onmessage) {
    target.onmessage(e);
  }

  target.dispatchEvent(e);
};

class ServiceWorker extends EventTarget {

  /**
   * Create a custom worker.
   */
  constructor() {
    super();

    // Service worker context.
    // Ie., A fake of ServiceWorkerGlobalScope.
    // The variable `globalScope` is equivalent to the `self` field of the Service Worker.
    // It will be used by the proauth.service.Controller as the ServiceWorker's global scope object.
    const globalScope = new EventTarget();

    const self = this;
    globalScope.clients = {
      matchAll() {
        return new Promise((resolve) => {
          resolve([{
            postMessage: (...args) => postMessage(self, ...args)
          }]);
        })
      }
    };

    // Client context
    // Used to send message from the client to the service worker
    this.controller = new class {
      postMessage(...args) {
        return postMessage(globalScope, ...args)
      }
    }();

    this.self = globalScope;

  }
}

class OAuth2Response {

  /**
   * Create an OAuth2.0 response.
   *
   * @param {Response} data The response
   */
  constructor(data) {
    this.accessToken = data.access_token;
    this.refreshToken = data.refresh_token;
    this.expiresIn = data.expires_in;
    this.tokenType = data.token_type;
    this.scope = data.scope;
  }

  toObject() {
    return Object.assign({}, this)
  }

}

class OAuth2Error {

  /**
   * Create an OAuth2.0 error.
   *
   * @param {Response} response The response
   */
  constructor(response) {
    this.response = response;
  }

  get status() {
    return this.response.status || null
  }

  get title() {
    return this.response.title || null
  }

  get detail() {
    return this.response.detail || null
  }

}

class OAuth2Client {

  /**
   * Create an OAuth2.0 client instance.
   *
   * @param {string} apiUrl The url of OAuth2 endpoint
   * @param {string} clientId A string identifying the instance of the client
   * @param {Function} fetch The fetch function the client will use
   */
  constructor(apiUrl, clientId, fetch) {

    /** @type {string} */
    this.apiUrl = apiUrl;

    /** @type {string} */
    this.clientId = clientId;

    this.fetch = (init) => {
      return fetch(this.apiUrl, init)
    };

  }

  handleAuthenticationResponse(response) {
    if (response.ok) {
      // Authentication OK
      return response.json().then((j) => new OAuth2Response(j))
    }

    // Authentication refused
    return response.json().then((j) => Promise.reject(new OAuth2Error(j)))
  }

  /**
   * Grant an access token.
   *
   * @param {Object} params An object containing the parameters to post to the OAuth2 endpoint
   * @return {Promise} A promise
   * @private
   */
  grantAccessToken(params) {
    return this.fetch({
      method: "POST",
      headers: {
        "Content-Type": "application/json",
        "Accept": "application/json"
      },
      body: JSON.stringify(params)
    })
  }

  /**
   * Grant a token for an user.
   *
   * @param {String} username The user for which to request a token
   * @param {String} password The password to authenticate such user
   * @return {Promise} A promise
   */
  userCredentials(username, password) {
    return this.grantAccessToken({
      "client_id": this.clientId,
      "grant_type": "password",
      username,
      password
    });
  }

  /**
   * Refresh an overdue token.
   *
   * @param {String} refreshToken The token to refresh
   * @return {Promise} A promise
   */
  refreshToken(refreshToken) {
    return this.grantAccessToken({
      "client_id": this.clientId,
      "grant_type": "refresh_token",
      "refresh_token": refreshToken
    });
  }

}

const attachAuthorization = (session, request) => {
  const tokens = session.content || {};

  if (tokens.accessToken) {

    /*
     * We have to create a new request based on the old one because the headers of the
     * original request are immutable in ServiceWorker.
     * Note that this is unnecessary if the code doesn't run in the ServiceWorker.
     * TODO: search for a better way to achieve the same result.
     */
    request = new Request(request);

    request.headers.set("Authorization", `${tokens.tokenType} ${tokens.accessToken}`);
  }

  return request
};

const updateSession = (session, grantPromise) => {
  return grantPromise.then((oA2Response) => {
    session.content = oA2Response.toObject();
    return true
  }, (oA2Error) => {
    session.clear();
    return false
  })
};

class Oauth2Handler {

  /**
   * Create an OAuth2.0 handler.
   *
   * @param {Object} settings The settings
   * @param {SessionHandler} session The session
   * @param {Function} fetch The fetch function
   */
  constructor(settings, session, fetch) {
    this.settings = settings;
    this.session = session;
    this.fetch = fetch;
    this.oauth2Client = new OAuth2Client(settings.oauthUrl, settings.oauthClientId, fetch);
    this.invalidGrantHttpCode = 401;
  }

  /**
   * Update the session on response
   *
   * @param {FetchEvent} event The intercepted event
   * @returns {void}
   */
  handleAuthentication(event) {
    const fetch = this.fetch;
    event.respondWith(fetch(event.request).then((response) => {
      // Just update the session on response
      return updateSession(
        this.session,
        this.oauth2Client.handleAuthenticationResponse(response.clone())
      ).then(() => response)
    }));
  }

  /**
   * Set the current token and handle 401s
   *
   * @param {FetchEvent} event The intercepted event
   * @returns {void}
   */
  handle(event) {

    const { oauthUrl } = this.settings;
    let request = event.request;
    const url = request.url;
    const fetch = this.fetch;

    // Is it an auth request?
    if (url === oauthUrl || `${url}/` === oauthUrl) {
      this.handleAuthentication(event);
      return
    }

    if (this._pendingRefresh) {
      // Wait for the fresh token
      this._pendingRefresh.then(() => {
        request = attachAuthorization(this.session, request);
        event.respondWith(fetch(request));
      });
      return
    }

    // Attach the current token
    request = attachAuthorization(this.session, request);

    // Finally, fetch and handle the response
    let requestCopy = request.clone();
    event.respondWith(fetch(request).then((response) => {

      if (this._pendingRefresh) {
        // Wait for the fresh token
        return this._pendingRefresh.then(() => {
          requestCopy = attachAuthorization(this.session, requestCopy);
          return fetch(requestCopy)
        })
      }

      const tokens = this.session.content || {};
      if (response.status === this.invalidGrantHttpCode && tokens.refreshToken) {
        // Got 401, but we have a refresh token, so try to get a new access token

        const resetPending = (param) => {
          this._pendingRefresh = null;
          return param
        };

        const oauth = this.oauth2Client;
        this._pendingRefresh = oauth.refreshToken(tokens.refreshToken).then(
          resetPending,
          resetPending
        ).then((authResponse) => {
          return updateSession(
            this.session,
            oauth.handleAuthenticationResponse(authResponse)
          ).then((newSession) => {
            // We got a new token?
            // Then try to re-fetch the original request...
            if (newSession) {
              requestCopy = attachAuthorization(this.session, requestCopy);
              return fetch(requestCopy)
            } //...else passthrough the 401 response
            return response
          })
        });
        return this._pendingRefresh
      }
      return response // Passthrough any others
    }));
  }

}

class SessionContainer {

  /**
   * Create a session container.
   *
   * @param {String} namespace The namespace
   * @param {Function} notify The notifier
   * @param {Object} data The content
   */
  constructor(namespace, notify, data) {
    this.notify = () => notify(namespace);
    this._content = data ? Object(data) : null;
  }

  get content() {
    return this._content
  }

  set content(data) {
    this._content = data ? Object(data) : null;
    this.notify();
  }

  clear() {
    this.content = null;
  }

}

class Controller {

  /**
   * Create a controller.
   *
   * @param {EventTarget} globalScope The Service Worker's global scope
   * @param {Function} fetch The fetch function it will use
   */
  constructor(globalScope, fetch) {

    this.globalScope = globalScope;

    if (!fetch) {
      fetch = globalScope.fetch;
    }

    this.handlers = {};

    // Setup interceptor
    globalScope.addEventListener("fetch", (event) => {
      for (const ns in this.handlers) {
        const handler = this.handlers[ns];
        for (const url of handler.settings.managedUrls) {
          if (event.request.url.startsWith(url)) {
            // Each managed url should be handle by just one handler,
            // further registered handler on the same url are discarded.
            // (todo): Should be a warning/error thrown in case of multiple handlers?
            handler.handle(event);
            return
          }
        }
      }

      // Passthrough unmanaged requests
      event.respondWith(fetch(event.request));
    });


    globalScope.addEventListener("message", (e) => {
      const { broadcast, namespace, command, params } = e.data;
      const reply = (...args) => e.ports[0].postMessage(...args);

      if (broadcast) {
        return // Discard broadcast messages
      }

      try {
        switch (command) {
        case "init":
          ((settings, sessionData = {}) => {
            const ns = settings.namespace;
            let h = this.handlers[ns];

            if (namespace !== ns) {
              throw Error("Invalid settings: namespace mismatch")
            }

            // Setup or override handler settings and session data
            if (h) {
              h = this.getHandlerByNamespace(ns);
              h.settings = settings;
              h.session.content = sessionData;
            } else {
              const s = new SessionContainer(ns, this.notifySession.bind(this), sessionData);
              h = this.handlers[ns] = new Oauth2Handler(settings, s, fetch);
            }

          })(...params);
          reply(true);
          break;

        case "setSession":
          ((data) => {
            this.getHandlerByNamespace(namespace).session.content = data;
          })(...params);
          reply(true);
          break;

        default:
          break;
        }
      } catch (error) {
        reply({
          namespace,
          error: {
            message: String(error.message),
            stack: String(error.stack)
          }
        });
      }
    });
  }

  getHandlerByNamespace(namespace) {
    if (!this.handlers[namespace]) {
      throw Error(`No handler found for '${namespace}'`)
    }
    return this.handlers[namespace]
  }

  sendMessage(namespace, command, ...params) {
    this.globalScope.clients.matchAll().then((clients) => {
      for (const client of clients) {
        client.postMessage({
          broadcast: true, namespace, command, params
        });
      }
    });
  }

  notifySession(namespace) {
    this.sendMessage(namespace, "session", this.getHandlerByNamespace(namespace).session.content);
  }
}

class SessionContainer$1 {

  /**
   * Create a session container.
   *
   * @param {String} namespace The namespace
   * @param {Storage} storage The session storage
   */
  constructor(namespace, storage) {
    this.namespace = namespace;
    this.storage = storage;
  }

  get content() {
    try {
      return JSON.parse(this.storage.getItem(this.namespace))
    } catch (error) {
    }
    return null
  }

  set content(data) {
    if (data) {
      this.storage.setItem(this.namespace, JSON.stringify(data || null));
    } else {
      this.clear();
    }
  }

  clear() {
    this.storage.removeItem(this.namespace);
  }

}

const sendMessage = (serviceWorker, message) => {
  // This wraps the message posting/response in a promise,
  // which will resolve if the response doesn't
  // contain an error, and reject with the error if it does.
  // If you'd prefer, it's possible to call
  // controller.postMessage() and set up the onmessage handler
  // independently of a promise, but this is a convenient wrapper.
  return new Promise((resolve, reject) => {
    const messageChannel = new MessageChannel();
    messageChannel.port1.onmessage = (event) => {
      if (event.data.error) {
        reject(event.data.error);
      } else {
        resolve(event.data);
      }
    };

    // This sends the message data as well as transferring messageChannel.port2
    // to the service worker.
    // The service worker can then use the transferred port to reply via postMessage(),
    // which will in turn trigger the onmessage handler on messageChannel.port1.
    // See https://html.spec.whatwg.org/multipage/workers.html#dom-worker-postmessage.
    serviceWorker.controller.postMessage(message, [messageChannel.port2]);
  })
};

class Client {

  /**
   * @typedef {Object} ClientSettings
   * @property {String} namespace
   * @property {String} oauthUrl
   * @property {String} oauthClientId
   * @property {String} sessionStorage
   */

  /**
   * Create a client.
   *
   * @param {ClientSettings} settings An object containing the client settings
   */
  constructor(settings) {
    const { namespace, sessionStorage } = settings;
    this.settings = settings;
    this.sessionContainer = new SessionContainer$1(namespace, window[sessionStorage]);
    this._onMessage = (event) => {
      const { broadcast, namespace: ns, command, params } = event.data;
      if (!broadcast || ns !== this.settings.namespace) {
        return
      }

      switch (command) {
        case "session":
          this.sessionContainer.content = params[0];
          break;

        default:
      }

      // (todo): set session race condition must be avoided, dirty-checking data may be a solution
    };

    this._whenReady = new Promise((resolve) => {
      this._resolveWhenReady = resolve;
    });
  }

  get ready() {
    return this._serviceWorker && this._ready
  }

  get whenReady() {
    return this._whenReady
  }

  get legacyMode() {
    return this._legacyMode
  }

  set legacyMode(sw) {
    throw Error("legacyMode is a read-only property")
  }

  get serviceWorker() {
    if (!this._serviceWorker) {
      throw Error("proauth.js: serviceWorker has been not registered yet.")
    }
    return this._serviceWorker;
  }

  set serviceWorker(sw) {
    if (this._serviceWorker) {
      this._serviceWorker.removeEventListener("message", this._onMessage);
    }

    this._ready = false;
    this._serviceWorker = sw;
    this._legacyMode = sw !== navigator.serviceWorker;

    sendMessage(sw, {
      namespace: this.settings.namespace,
      command: "init",
      params: [this.settings, this.sessionContainer.content]
    }).then((data) => {
      this._serviceWorker.addEventListener("message", this._onMessage);
      this._ready = !!data;
      if (this._ready) {
        this._resolveWhenReady();
      }
    });
  }

  setSession(data) {
    return sendMessage(this.serviceWorker, {
      namespace: this.settings.namespace,
      command: "setSession",
      params: [data]
    })
  }

  clearSession() {
    return this.setSession(null)
  }

  hasSession() {
    return !!this.sessionContainer.content;
  }

  login(user, pass) {
    if (!this.ready) {
      throw Error("proauth.js is not ready yet.")
    }
    const { oauthUrl, oauthClientId } = this.settings;
    return (new OAuth2Client(oauthUrl, oauthClientId, window.fetch))
      .userCredentials(user, pass)
      .then((response) => response.ok ? response : Promise.reject(response))
  }

}

const boot = () => {
  const settings = {
    // Put defaults here
    swSrc: "/service-worker.js",
    swOptions: {},
    oauthUrl: "/oauth",
    oauthClientId: "proauth",
    sessionStorage: "localStorage",
    namespace: document.origin && document.origin !== "null" ? document.origin : "",
    managedUrls: [
      "/" // Manage the current url root by default
    ]
  };

  // Copy user settings
  Object.assign(settings, window.proauth || {});

  // Patch settings
  settings.oauthUrl = absolutePath(settings.oauthUrl);
  for (const k in settings.managedUrls) {
    settings.managedUrls[k] = absolutePath(settings.managedUrls[k]);
  }

  const client = new Client(settings);

  return client;
};

var proauthClient = {
  client: boot()
};

const client = proauthClient.client;

// Init elements for legacy mode
const serviceWorker = new ServiceWorker(); // Fake serviceWorker
fetch$2.install(serviceWorker.self); // FetchAPI proxy

xhr.install(); // XMLHttpRequest to FetchAPI proxy

// Setup service environment
const originalFetch = fetch$2.fetch;

const effectiveFetch = originalFetch.polyfill ?
  (...args) => {
    xhr.uninstall();
    const ret = originalFetch(...args);
    xhr.install();
    return ret
  } :
  originalFetch;


const controller = new Controller(serviceWorker.self, effectiveFetch);
client.serviceWorker = serviceWorker;

const legacy = { fetch: fetch$2, xhr, serviceWorker };
const service = { controller };
var legacy$1 = { legacy, service, client };

return legacy$1;

}());
//# sourceMappingURL=legacy.es2015.js.map
